var helper = require("../../test-helper")
var buster = require("buster");
buster.remoteRunner = helper.require("test-runner/remote-runner");
var assert = buster.assert;

buster.testCase("Remote runner", {
    setUp: function () {
        this.clock = this.useFakeTimers();
        this.multicaster = buster.eventEmitter.create();

        this.emit = function (event, data, clientId, client) {
            return this.multicaster.emit(event, {
                topic: event,
                data: data,
                clientId: clientId,
                client: client
            });
        };

        this.runner = buster.remoteRunner.create(this.multicaster, { config: 42 });

        this.subscribeTo = function (event) {
            var listener = this.spy();
            this.runner.on(event, listener);

            return listener;
        };

        this.subscribeToMany = function () {
            var listeners = [];

            for (var i = 0, l = arguments.length; i < l; ++i) {
                listeners.push(this.subscribeTo(arguments[i]));
            }

            return listeners;
        };

        this.clients = ["Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.2.12) Gecko/20101027 Ubuntu/10.10 (maverick) Firefox/4.0",
                        "Mozilla/5.0 (X11; U; Linux i686; en-US) AppleWebKit/534.7 (KHTML, like Gecko) Chrome/11.0.517.44 Safari/534.7"];
    },

    "starting a run": {
        "should emit tests:run when client emits ready": function () {
            var client = buster.eventEmitter.create();
            var listener = this.spy();
            client.on("tests:run", listener);

            this.emit("ready", this.clients[0], 1, client);

            assert.calledOnce(listener);
        },

        "should emit tests:run with config": function () {
            var client = buster.eventEmitter.create();
            var listener = this.spy();
            client.on("tests:run", listener);

            this.emit("ready", this.clients[0], 1, client);

            assert.calledWith(listener, { config: 42 });
        },

        "should set client info when client is ready": function () {
            this.emit("ready", this.clients[0], 1, buster.eventEmitter.create());

            assert.match(this.runner.getClient(1), {
                browser: "Firefox",
                version: "4.0",
                platform: "Linux"
            });
        },

        "should enumerate client if a similar one exists": function () {
            this.emit("ready", this.clients[0], 1, buster.eventEmitter.create());
            this.emit("ready", this.clients[0], 2, buster.eventEmitter.create());

            assert.equals(this.runner.getClient(1).toString(), "Firefox 4.0 Linux");
            assert.equals(this.runner.getClient(2).toString(), "Firefox 4.0 Linux (2)");
        },

        "should keep enumerating clients from same browser": function () {
            this.emit("ready", this.clients[0], 1, buster.eventEmitter.create());
            this.emit("ready", this.clients[0], 2, buster.eventEmitter.create());
            this.emit("ready", this.clients[0], 3, buster.eventEmitter.create());
            this.emit("ready", this.clients[0], 4, buster.eventEmitter.create());

            assert.equals(this.runner.getClient(4).toString(), "Firefox 4.0 Linux (4)");
        },

        "should emit client:connect when client emits ready": function () {
            var listener = this.spy();
            this.runner.on("client:connect", listener);

            this.emit("ready", this.clients[1], 2, buster.eventEmitter.create());

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                browser: "Chrome",
                version: "11.0",
                platform: "Linux"
            });
        }
    },

    "while tests are running": {
        setUp: function () {
            this.emit("ready", this.clients[0], 1, buster.eventEmitter.create());
            this.emit("ready", this.clients[1], 2, buster.eventEmitter.create());
        },

        "should emit progress:suite:start event": function () {
            var listener = this.subscribeTo("progress:suite:start");
            this.emit("suite:start", null, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" }
            });
        },

        "should emit progress:suite:start event for all clients": function () {
            var listener = this.subscribeTo("progress:suite:start");
            this.emit("suite:start", null, 1);
            this.emit("suite:start", null, 2);

            assert.calledTwice(listener);
            assert.match(listener.args[0][0], {client: {id: 1, browser: "Firefox"}});
            assert.match(listener.args[1][0], {client: {id: 2, browser: "Chrome"}});
        },

        "should emit progress:suite:end event": function () {
            var listener = this.subscribeTo("progress:suite:end");
            this.emit("suite:end", null, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" }
            });
        },

        "should emit progress:suite:end event for all clients": function () {
            var listener = this.subscribeTo("progress:suite:end");
            this.emit("suite:end", null, 1);
            this.emit("suite:end", null, 2);

            assert.calledTwice(listener);
            assert.match(listener.args[0][0], {client: {id: 1, browser: "Firefox"}});
            assert.match(listener.args[1][0], {client: {id: 2, browser: "Chrome"}});
        },

        "should emit progress:test:success event": function () {
            var listener = this.subscribeTo("progress:test:success");
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:success", { name: "test #1" }, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" },
                name: "test #1",
                contexts: ["Test case"]
            });
        },

        "should emit progress:test:success event for different clients": function () {
            var listener = this.subscribeTo("progress:test:success");
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("context:start", { name: "Test case" }, 2);
            this.emit("test:success", { name: "test #1" }, 2);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 2, browser: "Chrome" },
                name: "test #1"
            });

            assert.equals(listener.args[0][0].contexts, ["Test case"]);
        },

        "should emit progress:test:error event": function () {
            var listener = this.subscribeTo("progress:test:error");
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:error", { name: "test #1", error: { message: "!" } }, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" },
                name: "test #1",
                contexts: ["Test case"],
                error: { message: "!" }
            });
        },

        "should emit progress:test:failure event": function () {
            var listener = this.subscribeTo("progress:test:failure");
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:failure", { name: "test #1", error: { message: "!" } }, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" },
                name: "test #1",
                contexts: ["Test case"],
                error: { message: "!" }
            });
        },

        "should emit progress:test:failure event for nested test": function () {
            var listener = this.subscribeTo("progress:test:failure");
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("context:start", { name: "In situation" }, 1);
            this.emit("test:failure", { name: "test #1", error: { message: "!" } }, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" },
                name: "test #1",
                contexts: ["Test case", "In situation"],
                error: { message: "!" }
            });
        },

        "should emit progress:test:failure event for middle nested test": function () {
            var listener = this.subscribeTo("progress:test:failure");
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("context:start", { name: "In situation" }, 1);
            this.emit("context:end", { name: "In situation" }, 1);
            this.emit("context:start", { name: "Other" }, 1);
            this.emit("test:failure", { name: "test #1", error: { message: "!" } }, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" },
                name: "test #1",
                contexts: ["Test case", "Other"],
                error: { message: "!" }
            });
        },

        "should emit progress:test:timeout event": function () {
            var listener = this.subscribeTo("progress:test:timeout");
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:timeout", { name: "test #1" }, 1);

            assert.calledOnce(listener);
            assert.match(listener.args[0][0], {
                client: { id: 1, browser: "Firefox" },
                name: "test #1",
                contexts: ["Test case"]
            });
        }
    },

    "when context is completed": {
        setUp: function () {
            this.emit("ready", this.clients[0], 1, buster.eventEmitter.create());
            this.emit("ready", this.clients[1], 2, buster.eventEmitter.create());
        },

        "should emit test runner events": function () {
            var listeners = this.subscribeToMany(
                "context:start", "context:end", "test:success",
                "test:setUp", "test:tearDown", "test:start");

            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:setUp", { name: "test #1" }, 1);
            this.emit("test:start", { name: "test #1" }, 1);
            this.emit("test:tearDown", { name: "test #1" }, 1);
            this.emit("test:success", { name: "test #1" }, 1);

            assert.notCalled(listeners[0]);

            this.emit("context:end", { name: "Test case" }, 1);

            assert.calledTwice(listeners[0]);
            assert.equals(listeners[0].args[0][0], { name: "Firefox 4.0 Linux" });
            assert.equals(listeners[0].args[1][0], { name: "Test case" });
            assert.calledTwice(listeners[1]);
            assert.equals(listeners[1].args[0][0], { name: "Test case" });
            assert.equals(listeners[1].args[1][0], { name: "Firefox 4.0 Linux" });
            assert.calledOnce(listeners[2]);
            assert.calledWith(listeners[2], { name: "test #1" });
            assert.calledOnce(listeners[3]);
            assert.calledWith(listeners[3], { name: "test #1" });
            assert.calledOnce(listeners[4]);
            assert.calledWith(listeners[4], { name: "test #1" });
            assert.calledOnce(listeners[5]);
            assert.calledWith(listeners[5], { name: "test #1" });
        },

        "should emit log through console property": function () {
            var logger = { log: this.spy() };
            this.runner.console.bind(logger, "log");

            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("log", { level: "log", message: "Hey" }, 1);
            this.emit("context:end", { name: "Test case" }, 1);

            assert.calledOnce(logger.log);
            assert.calledWith(logger.log, { level: "log", message: "Hey" });
        },

        "should emit context:unsupported event": function () {
            var listener = this.spy();
            this.runner.on("context:unsupported", listener);

            this.emit("context:unsupported", { name: "Test case" }, 1);

            assert.calledOnce(listener);
        },

        "should not emit context:unsupported event while in nested context": function () {
            var listener = this.spy();
            this.runner.on("context:unsupported", listener);

            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("context:unsupported", { name: "Something" }, 1);
            assert.notCalled(listener);

            this.emit("context:end", { name: "Test case" }, 1);
            assert.calledOnce(listener);
        },

        "should emit test runner test status events": function () {
            var listeners = this.subscribeToMany(
                "test:error", "test:failure", "test:timeout", "test:deferred");

            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:error", { name: "test #1" }, 1);
            this.emit("test:failure", { name: "test #2" }, 1);
            this.emit("test:timeout", { name: "test #3" }, 1);
            this.emit("test:deferred", { name: "test #4" }, 1);
            this.emit("context:end", { name: "Test case" }, 1);

            assert.calledOnce(listeners[0]);
            assert.calledWith(listeners[0], { name: "test #1" });
            assert.calledOnce(listeners[1]);
            assert.calledWith(listeners[1], { name: "test #2" });
            assert.calledOnce(listeners[2]);
            assert.calledWith(listeners[2], { name: "test #3" });
            assert.calledOnce(listeners[3]);
            assert.calledWith(listeners[3], { name: "test #4" });
        },

        "should not re-emit old test runner events": function () {
            var listener = this.subscribeTo("test:error");

            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:error", { name: "test #1" }, 1);
            this.emit("context:end", { name: "Test case" }, 1);
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:failure", { name: "test #2" }, 1);
            this.emit("context:end", { name: "Test case" }, 1);

            assert.calledOnce(listener);
        },

        "should not emit until full context is ready": function () {
            var listener = this.subscribeTo("test:error");

            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:error", { name: "test #1" }, 1);
            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:failure", { name: "test #2" }, 1);
            this.emit("context:end", { name: "Test case" }, 1);

            assert.notCalled(listener);
            this.emit("context:end", { name: "Test case" }, 1);

            assert.calledOnce(listener);
        },

        "should emit context start/stop in correct order": function () {
            var out = [];

            this.runner.on("context:start", function (ctx) {
                out.push("START " + ctx.name);
            });

            this.runner.on("context:end", function (ctx) {
                out.push("STOP " + ctx.name);
            });

            this.runner.on("test:success", function (test) {
                out.push("YAY " + test.name);
            });


            this.emit("context:start", { name: "Test case" }, 1);
            this.emit("test:success", { name: "test #1" }, 1);
            this.emit("context:start", { name: "Inner" }, 1);
            this.emit("test:success", { name: "test #2" }, 1);
            this.emit("context:start", { name: "Inner inner" }, 1);
            this.emit("test:success", { name: "test #3" }, 1);
            this.emit("context:end", { name: "Inner inner" }, 1);
            this.emit("context:end", { name: "Inner" }, 1);
            this.emit("context:start", { name: "Inner #2" }, 1);
            this.emit("test:success", { name: "test #4" }, 1);
            this.emit("context:end", { name: "Inner #2" }, 1);
            this.emit("context:end", { name: "Test case" }, 1);

            assert.equals(out, [
                "START Firefox 4.0 Linux",
                "START Test case", "YAY test #1",
                "START Inner", "YAY test #2",
                "START Inner inner", "YAY test #3", "STOP Inner inner",
                "STOP Inner",
                "START Inner #2", "YAY test #4", "STOP Inner #2",
                "STOP Test case",
                "STOP Firefox 4.0 Linux"
            ]);
        }
    },

    "test suites": {
        setUp: function () {
            this.emit("ready", this.clients[0], 1, buster.eventEmitter.create());
            this.emit("ready", this.clients[1], 2, buster.eventEmitter.create());
        },

        "should emit suite:start only once": function () {
            var listener = this.subscribeTo("suite:start");

            this.emit("suite:start", {}, 1);
            this.emit("suite:start", {}, 2);

            assert.calledOnce(listener);
        },

        "should not emit suite:end when clients are still in progress": function () {
            var listener = this.subscribeTo("suite:end");

            this.emit("suite:start", {}, 1);
            this.emit("suite:start", {}, 2);
            this.emit("suite:end", {}, 1);

            assert.notCalled(listener);
        },

        "should emit suite:end when all clients are ready": function () {
            var listener = this.subscribeTo("suite:end");

            this.emit("suite:start", {}, 1);
            this.emit("suite:start", {}, 2);
            this.emit("suite:end", {}, 2);
            this.emit("suite:end", {}, 1);

            assert.calledOnce(listener);
        },

        "should not emit suite:end until ready - interleaved suites": function () {
            this.emit("ready", "", 3, buster.eventEmitter.create());
            this.emit("ready", "", 4, buster.eventEmitter.create());
            this.emit("ready", "", 5, buster.eventEmitter.create());

            var listener = this.subscribeTo("suite:end");

            this.emit("suite:start", {}, 1);
            this.emit("suite:start", {}, 2);
            this.emit("suite:end", {}, 2);
            this.emit("suite:start", {}, 3);
            this.emit("suite:start", {}, 4);
            this.emit("suite:end", {}, 1);
            this.emit("suite:end", {}, 3);
            this.emit("suite:start", {}, 5);
            this.emit("suite:end", {}, 4);

            assert.notCalled(listener);
            this.emit("suite:end", {}, 5);

            assert.calledOnce(listener);
        },

        "should not emit suite:end if some clients have not yet started": function () {
            var listener = this.subscribeTo("suite:end");

            this.emit("suite:start", {}, 1);
            this.emit("suite:end", {}, 1);
            assert.notCalled(listener);

            this.emit("suite:start", {}, 2);
            this.emit("suite:end", {}, 2);
            assert.calledOnce(listener);
        },

        "should summarize stats for suite:end": function () {
            var listener = this.subscribeTo("suite:end");

            this.emit("suite:start", {}, 1);
            this.emit("suite:start", {}, 2);

            this.emit("suite:end", {
                contexts: 5,
                tests: 10,
                errors: 2,
                failures: 3,
                assertions: 14,
                timeouts: 2,
                deferred: 1,
                ok: false
            }, 1);

            this.emit("suite:end", {
                contexts: 6,
                tests: 10,
                errors: 1,
                failures: 1,
                assertions: 4,
                timeouts: 1,
                deferred: 0,
                ok: false
            }, 2);

            assert.equals(listener.args[0][0], {
                clients: 2,
                contexts: 11,
                tests: 20,
                errors: 3,
                failures: 4,
                assertions: 18,
                timeouts: 3,
                deferred: 1,
                ok: false
            });
        }
    },

    "client timeouts": {
        setUp: function () {
            this.connect = function () {
                this.emit("ready", this.clients[0], 1, buster.eventEmitter.create());
                this.emit("ready", this.clients[1], 2, buster.eventEmitter.create());
            };
        },

        "should emit client:timeout if client is unresponsive for 15s": function () {
            this.connect(); 
            var listener = this.spy();
            this.runner.on("client:timeout", listener);

            this.clock.tick(14999);
            this.emit("context:start", { name: "Test" }, 1);
            assert.notCalled(listener);

            this.clock.tick(1);
            assert.calledOnce(listener);
            assert.match(listener.args[0][0], { browser: "Chrome" });
        },

        "should emit client:timeout if client is unresponsive for custom timeout": function () {
            this.runner.timeout = 2000;
            this.connect();
            var listener = this.spy();
            this.runner.on("client:timeout", listener);

            this.clock.tick(1999);
            this.emit("context:start", { name: "Test" }, 1);
            assert.notCalled(listener);

            this.clock.tick(1);
            assert.calledOnce(listener);
            assert.match(listener.args[0][0], { browser: "Chrome" });
        },

        "should not emit client:timeout after 15s if there has been activity": function () {
            this.connect();
            var listener = this.spy();
            this.runner.on("client:timeout", listener);

            this.clock.tick(7500);
            this.emit("context:start", { name: "Test" }, 1);
            this.emit("context:start", { name: "Test" }, 2);

            this.clock.tick(7500);
            assert.notCalled(listener);
        },

        "should complete run with only one client if other timed out": function () {
            this.connect();
            var listener = this.spy();
            this.runner.on("suite:end", listener);

            this.clock.tick(7500);
            this.emit("suite:start", null, 1);
            this.emit("context:start", { name: "Test" }, 1);
            this.emit("context:end", { name: "Test" }, 1);
            this.clock.tick(7500);
            this.emit("suite:end", null, 1);

            assert.calledOnce(listener);
        },

        "should clean up timers when suite completes": function () {
            this.connect();
            var listener = this.spy();
            this.runner.on("client:timeout", listener);

            this.emit("suite:start", null, 1);
            this.emit("suite:end", null, 1);
            this.emit("suite:start", null, 2);
            this.emit("suite:end", null, 2);

            this.clock.tick(15000);

            assert.notCalled(listener);
        },

        "should ignore suite messages from timed out client": function () {
            this.connect();
            var listener = this.spy();
            this.runner.on("suite:start", listener);
            this.runner.on("progress:suite:start", listener);
            this.runner.on("progress:suite:end", listener);
            this.runner.on("suite:end", listener);

            this.clock.tick(15000);
            this.emit("suite:start", null, 1);
            this.emit("suite:end", null, 1);

            assert.notCalled(listener);
        },

        "should ignore context messages from timed out client": function () {
            this.connect();
            var listener = this.spy();
            this.runner.on("context:start", listener);
            this.runner.on("context:unsupported", listener);
            this.runner.on("context:end", listener);

            this.clock.tick(15000);
            this.emit("context:start", { name: "Tests" }, 1);
            this.emit("context:unsupported", { name: "Tests" }, 1);
            this.emit("context:end", { name: "Tests" }, 1);

            assert.notCalled(listener);
        },

        "should ignore test messages from timed out client": function () {
            this.connect();
            var listener = this.spy();
            this.runner.on("progress:test:success", listener);
            this.runner.on("progress:test:error", listener);
            this.runner.on("progress:test:failure", listener);
            this.runner.on("progress:test:timeout", listener);
            this.runner.on("log", listener);
            this.runner.on("test:start", listener);
            this.runner.on("test:setUp", listener);
            this.runner.on("test:tearDown", listener);
            this.runner.on("test:success", listener);
            this.runner.on("test:error", listener);
            this.runner.on("test:failure", listener);
            this.runner.on("test:timeout", listener);
            this.runner.on("test:deferred", listener);

            this.clock.tick(15000);
            this.emit("context:start", { name: "Tests" }, 1);
            this.emit("test:start", { name: "#1" }, 1);
            this.emit("test:setUp", { name: "#1" }, 1);
            this.emit("test:success", { name: "#1" }, 1);
            this.emit("test:error", { name: "#1" }, 1);
            this.emit("test:failure", { name: "#1" }, 1);
            this.emit("test:timeout", { name: "#1" }, 1);
            this.emit("test:deferred", { name: "#1" }, 1);
            this.emit("log", { name: "#1" }, 1);
            this.emit("context:end", { name: "Tests" }, 1);

            assert.notCalled(listener);
        }
    }
});
